/*
 * Copyright (C) 2017-2022 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <assert.h>
#include "proba-flooding-ring-routing-agent.h"
#include "../scheduler.h"

//==============================================================================
//
//          ProbaFloodingRingRoutingAgent  (class)
//
//==============================================================================

mt19937_64 *ProbaFloodingRingRoutingAgent::forwardingRNG =  new mt19937_64(2);
set<int> ProbaFloodingRingRoutingAgent::reachability = set<int>();

ProbaFloodingRingRoutingAgent::ProbaFloodingRingRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  alreadySent1and2 = false;
  rangeBig = ScenarioParameters::getRangeBig();
  rangeSmall = ScenarioParameters::getRangeSmall();
  //      aliveAgents++;
  forwardingProbability = (ScenarioParameters::getProbability()) * hostNode->getNeighboursCount() / hostNode->getRingNeighboursCount(ScenarioParameters::getRangeSmall(),ScenarioParameters::getRangeBig());
  //cout << "toto " << forwardingProbability << endl;
}

ProbaFloodingRingRoutingAgent::~ProbaFloodingRingRoutingAgent() {
  aliveAgents--;
  //   cout << " aliveAgents " << aliveAgents << endl;

  if ( aliveAgents == 0 ){
    //           TODO : add global information
    cout << " Proba flooding with ring:: number of nodes reached : " << reachability.size() << endl;
    cout << " Proba flooding with ring:: forwarded packets       : " << forwardedDataPackets << endl;
  }
}

void ProbaFloodingRingRoutingAgent::initializeAgent() {
  readRangeAttributes ();
  readProbaAttribute ();
}

void ProbaFloodingRingRoutingAgent::receivePacketFromApplication(PacketPtr _packet) { //on source

  if (!alreadySent1and2) { //1 and 2 before the very first data transmission
    PacketPtr control1(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
    //control1->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber()); 
    PacketPtr control2(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
    //control2->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());

    control1->setCommunicationRange(rangeBig);
    hostNode->enqueueOutgoingPacket(control1);
    control2->setCommunicationRange(rangeSmall);
    hostNode->enqueueOutgoingPacket(control2);
		
    alreadySent1and2 = true;	
  }

  _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
  hostNode->enqueueOutgoingPacket(_packet);   
}


void ProbaFloodingRingRoutingAgent::receivePacketFromNetwork(PacketPtr _packet) {
  // source must not resend the packet
  if ( _packet->srcId == hostNode->getId())
    return;

  bool fw = false;
  uniform_real_distribution<float> distrib(0, 1);
  float forwardingProbabilityNumber = distrib(*forwardingRNG);

  if (_packet->type == PacketType::DATA){
    
    /////////
    
    auto transmitterIt = neighb.find(_packet->transmitterId);
    
    
    if ( transmitterIt == neighb.end() ) {
      cout << "pas trouvé transmitterId " << _packet->transmitterId << " dans neighb" << endl;
    }
    //assert(transmitterIt != neighb.end());
      
    bool isOnTransmitterRing = transmitterIt->second.control1 && !transmitterIt->second.control2;

    auto sourceRingIt = sourceRing.find(_packet->srcId);
     
    if ( sourceRingIt == sourceRing.end() ) {
      sourceRing.insert(pair<int,bool>(_packet->srcId,isOnTransmitterRing));
    } else {
      if ( !isOnTransmitterRing ) {
        sourceRingIt->second = false;
      }
    }
    /////////
    
    auto srcIt = alreadySeenPackets.find(_packet->srcId);
    if (srcIt != alreadySeenPackets.end()) {
      // already seen
      if (_packet->srcSequenceNumber > srcIt->second) {
	// the source sequence number of this packet is higher than the highest already seen from this source
	srcIt->second = _packet->srcSequenceNumber;
	fw = true;
      }
    } else {
      // seen for the first time
      alreadySeenPackets.insert(pair<int,int>(_packet->srcId, _packet->srcSequenceNumber));
      fw = true;
		
    }
    if (fw && forwardingProbabilityNumber < forwardingProbability && _packet->type == PacketType::DATA) {
      processPacketAndFwIfOnRing(_packet);
    }
  }

  else if (_packet->type == PacketType::CONTROL_1 || _packet->type == PacketType::CONTROL_2) {
    processPacketAndFwIfOnRing(_packet);

  }


  // packet arrives to destination (the destination can be unicast, zone
  // (SLR), or any node if broadcast to all network)
  // note that all the copies are sent upwards to application, it is up to
  // it to process them only once
  // note also that the destination too forwards the packet, useful in SLR
  // for ex., where all the zone should receive the packet
  if (_packet->dstId == hostNode->getId() || _packet->dstId == -1) {
    hostNode->dispatchPacketToApplication(_packet);
    reachability.insert(hostNode->getId());
    LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber);
  }
}



void ProbaFloodingRingRoutingAgent::processPacketAndFwIfOnRing( PacketPtr _packet ) {

  auto nodeIt = neighb.find( _packet->transmitterId );
  if (nodeIt == neighb.end()) {
    // first time we see a packet coming from this node	
    Controlll12_t control;
    control.control1 = false;
    control.control2 = false;

    if ( _packet->type == PacketType::CONTROL_1 ) control.control1 = true;
    if ( _packet->type == PacketType::CONTROL_2 ) control.control2 = true;

    neighb.insert(pair<int,Controlll12_t>(_packet->transmitterId, control));
    // Even if it is a DATA packet, we don't forward it as we did not previously see CONTROL1
  } else {
    // we have already seen a packet coming from this node
    if ( _packet->type == PacketType::CONTROL_1 ) {
      nodeIt->second.control1 = true;
    } else if ( _packet->type == PacketType::CONTROL_2 ) {
      nodeIt->second.control2 = true;
    } else if ( nodeIt->second.control1 && !nodeIt->second.control2 ) {
      if (!alreadySent1and2) { //1 and 2 before the very first data transmission
        PacketPtr control1(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));   
        PacketPtr control2(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));

        control1->setCommunicationRange(rangeBig);
        hostNode->enqueueOutgoingPacket(control1);

        control2->setCommunicationRange(rangeSmall);
        hostNode->enqueueOutgoingPacket(control2);
			
        alreadySent1and2 = true;	
      }

      PacketPtr packetClone(_packet->clone());
      
      if ( false ) {           
        // Don't use the delayed forwarding
 
        hostNode->enqueueOutgoingPacket(packetClone);
        if (packetClone->type == PacketType::DATA ){
          //cout << " je forward de la data " << endl;
          forwardedDataPackets++;
        }
      } else {

        auto sendEvent = new ProbaFloodingRingSendEvent(Scheduler::now()+200000000, this, packetClone);
        Scheduler::getScheduler().schedule(sendEvent);
        
        // Use the delayed forwarding
       // In the meantime, if it appears this not is not on the ring of another node, we discard the packet
      }
      
      
    }
  }

}

void ProbaFloodingRingRoutingAgent::processPacketSendEvent( PacketPtr _packet ) {

  auto srcIt = sourceRing.find( _packet->srcId );
  assert( srcIt != sourceRing.end() );
  
  if ( srcIt->second ) {
  
    hostNode->enqueueOutgoingPacket(_packet);
    if (_packet->type == PacketType::DATA ){
  
      //cout << " je forward de la data " << endl;
      forwardedDataPackets++;
    }
  }
}
