/*
 * Copyright (C) 2017-2022 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <cassert>
#include "raytracing-routing-agent.h"
#include "../scheduler.h"

//==============================================================================
//
//          RayTracingRoutingAgent  (class)
//
//==============================================================================

mt19937_64 *RayTracingRoutingAgent::backoffrng = new mt19937_64(0);

void RayTracingRoutingAgent::processSendEvent (PacketPtr packet) {
  int id = packet->srcSequenceNumber;  // TODO make id unique no matter the src/flow
  assert (id == 0);
  if (scheduled[id] == true) {
    hostNode->enqueueOutgoingPacket (packet);
    forwardedDataPackets++;
  }
}

RayTracingRoutingAgent::RayTracingRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  //      aliveAgents++;
  timefirst[0] = -1;
}

RayTracingRoutingAgent::~RayTracingRoutingAgent() {
  aliveAgents--;
  //   cout << " aliveAgents " << aliveAgents << endl;
  if ( aliveAgents == 0 )
    cout << " Linear fw:: forwarded packets: " << forwardedDataPackets << endl;
}


void RayTracingRoutingAgent::receivePacketFromApplication(PacketPtr packet) {
  packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
  packet->disableBackoff = true;
  hostNode->enqueueOutgoingPacket(packet);
}

void RayTracingRoutingAgent::receivePacketFromNetwork(PacketPtr packet) {
  //      cout << Scheduler::getScheduler().now() << hostNode->getId();
  //      cout << " srcId:" << _packet->srcId << " srcSeq:" << _packet->srcSequenceNumber;
  //      cout << " dstId:" << _packet->dstId << " port:" << _packet->port ;

  // source must not resend the packet
  if ( packet->srcId == hostNode->getId())
    return;

  /*static*/ simulationTime_t ts = Node::getPulseDuration() * packet->beta;
  int id = packet->srcSequenceNumber;  // TODO make id unique no matter the src/flow
  assert (id == 0);

  if (hostNode->getId() == 1) {  // steering node
    if (timefirst[id] == -1) {
      // transmit it at the 1st reception
      PacketPtr packetClone(packet->clone());
      hostNode->enqueueOutgoingPacket(packetClone);
      forwardedDataPackets++;
      timefirst[id] = 0;  // discard subsequent packets
    }  // else discard
    return;
  }

  if (timefirst[id] == -1) {  // packet id is seen for the first time
    timefirst[id] = Scheduler::now();
    second[id] = false;
  } else if (second[id] == false) { // 2nd packet id received
    second[id] = true;
    // cout << "test " << hostNode->getId() << " " << Scheduler::now()%ts << " " << timeslot[id] << " " << Node::getPulseDuration() << endl;
    if ((Scheduler::now() - timefirst[id])%ts < Node::getPulseDuration()
        || (Scheduler::now() - timefirst[id])%ts > ts - Node::getPulseDuration()) {  // collinearity
      //PacketPtr ppacket(new Packet(_packet));
      PacketPtr packetClone(packet->clone());
      // schedule event e[id] to forward p[id] at now+backoff
      uniform_int_distribution<simulationTime_t> distrib(0, 1000);  // TODO use a computed value, based on the density, instead of 1000
      // term at left of + is the backoff to allow nodes to hear other packets before sending, and term at right of + is the delay to have the same sending time modulo ts as the packet received
      simulationTime_t backoff = packet->size * ts * distrib(*backoffrng) + ts - Node::getPulseDuration();
      auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, packetClone);
      Scheduler::getScheduler().schedule (sendEvent);
      // if sending now:
      // hostNode->enqueueOutgoingPacket(packetClone);
      // forwardedDataPackets++;

      scheduled[id] = true;
    }
  } else {  // 3rd or subsequent packet
    if (scheduled[id] == true)
      scheduled[id] = false;  // unschedule event e[id], i.e. make it a noop
  }
}
